@@ -27,7 +27,7 @@
  *
  */
 // Windows only header file, guard it so that accidental inclusion doesn't cause unknown header include errors
-#ifdef _WIN32
+#if defined(_WIN32)
 
 // This needs to be defined first, or else we'll get redefinitions on NTSTATUS values
 #define UMDF_USING_NTSTATUS
@@ -45,13 +45,13 @@
 #include <devpkey.h>
 #include <winternl.h>
 #include <strsafe.h>
-#ifdef __MINGW32__
-#undef strcpy  // fix error with redfined strcpy when building with MinGW-w64
+#if defined(__MINGW32__)
+#undef strcpy  // fix error with redefined strcpy when building with MinGW-w64
 #endif
 #include <dxgi1_6.h>
 #include "adapters.h"
 
-#ifndef __MINGW32__
+#if !defined(__MINGW32__)
 // not yet available with MinGW-w64 stable
 #include <appmodel.h>
 #endif
@@ -61,7 +61,10 @@
 #endif
 
 typedef HRESULT(APIENTRY *PFN_CreateDXGIFactory1)(REFIID riid, void **ppFactory);
-static PFN_CreateDXGIFactory1 fpCreateDXGIFactory1;
+PFN_CreateDXGIFactory1 fpCreateDXGIFactory1;
+
+// Empty function just so windows_initialization can find the current module location
+void function_for_finding_the_current_module(void) {}
 
 void windows_initialization(void) {
     char dll_location[MAX_PATH];
@@ -69,20 +72,21 @@ void windows_initialization(void) {
 
     // Get a module handle to a static function inside of this source
     if (GetModuleHandleEx(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS | GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
-                          (LPCSTR)&loader_debug_init, &module_handle) != 0 &&
+                          (LPCSTR)&function_for_finding_the_current_module, &module_handle) != 0 &&
         GetModuleFileName(module_handle, dll_location, sizeof(dll_location)) != 0) {
         loader_log(NULL, VULKAN_LOADER_INFO_BIT, 0, "Using Vulkan Loader %s", dll_location);
     }
 
     // This is needed to ensure that newer APIs are available right away
     // and not after the first call that has been statically linked
-    LoadLibrary("gdi32.dll");
+    LoadLibraryEx("gdi32.dll", NULL, LOAD_LIBRARY_SEARCH_SYSTEM32);
 
     wchar_t systemPath[MAX_PATH] = L"";
     GetSystemDirectoryW(systemPath, MAX_PATH);
     StringCchCatW(systemPath, MAX_PATH, L"\\dxgi.dll");
     HMODULE dxgi_module = LoadLibraryW(systemPath);
-    fpCreateDXGIFactory1 = dxgi_module == NULL ? NULL : (PFN_CreateDXGIFactory1)GetProcAddress(dxgi_module, "CreateDXGIFactory1");
+    fpCreateDXGIFactory1 =
+        dxgi_module == NULL ? NULL : (PFN_CreateDXGIFactory1)(void *)GetProcAddress(dxgi_module, "CreateDXGIFactory1");
 
 #if !defined(NDEBUG)
     _set_error_mode(_OUT_TO_STDERR);
@@ -92,6 +96,7 @@ void windows_initialization(void) {
 }
 
 BOOL WINAPI DllMain(HINSTANCE hinst, DWORD reason, LPVOID reserved) {
+    (void)hinst;
     switch (reason) {
         case DLL_PROCESS_ATTACH:
             loader_initialize();
@@ -227,9 +232,9 @@ out:
 
 VkResult windows_get_device_registry_files(const struct loader_instance *inst, uint32_t log_target_flag, char **reg_data,
                                            PDWORD reg_data_size, LPCSTR value_name) {
-    static const wchar_t *softwareComponentGUID = L"{5c4c3332-344d-483c-8739-259e934c9cc8}";
-    static const wchar_t *displayGUID = L"{4d36e968-e325-11ce-bfc1-08002be10318}";
-#ifdef CM_GETIDLIST_FILTER_PRESENT
+    const wchar_t *softwareComponentGUID = L"{5c4c3332-344d-483c-8739-259e934c9cc8}";
+    const wchar_t *displayGUID = L"{4d36e968-e325-11ce-bfc1-08002be10318}";
+#if defined(CM_GETIDLIST_FILTER_PRESENT)
     const ULONG flags = CM_GETIDLIST_FILTER_CLASS | CM_GETIDLIST_FILTER_PRESENT;
 #else
     const ULONG flags = 0x300;
@@ -348,7 +353,7 @@ VkResult windows_get_registry_files(const struct loader_instance *inst, char *lo
                                     PDWORD reg_data_size) {
     // This list contains all of the allowed ICDs. This allows us to verify that a device is actually present from the vendor
     // specified. This does disallow other vendors, but any new driver should use the device-specific registries anyway.
-    static const struct {
+    const struct {
         const char *filename;
         unsigned int vendor_id;
     } known_drivers[] = {
@@ -553,7 +558,8 @@ VkResult windows_get_registry_files(const struct loader_instance *inst, char *lo
     }
 
     if (!found && result != VK_ERROR_OUT_OF_HOST_MEMORY) {
-        loader_log(inst, log_target_flag, 0, "Found no registry files in %s", location);
+        loader_log(inst, log_target_flag, 0, "Found no registry files in %s\\%s",
+                   (hive == DEFAULT_VK_REGISTRY_HIVE) ? DEFAULT_VK_REGISTRY_HIVE_STR : SECONDARY_VK_REGISTRY_HIVE_STR, location);
         result = VK_ERROR_INCOMPATIBLE_DRIVER;
     }
 
@@ -581,9 +587,10 @@ VkResult windows_read_manifest_from_d3d_adapters(const struct loader_instance *i
         goto out;
     }
 
-    PFN_LoaderEnumAdapters2 fpLoaderEnumAdapters2 = (PFN_LoaderEnumAdapters2)GetProcAddress(gdi32_dll, "D3DKMTEnumAdapters2");
+    PFN_LoaderEnumAdapters2 fpLoaderEnumAdapters2 =
+        (PFN_LoaderEnumAdapters2)(void *)GetProcAddress(gdi32_dll, "D3DKMTEnumAdapters2");
     PFN_LoaderQueryAdapterInfo fpLoaderQueryAdapterInfo =
-        (PFN_LoaderQueryAdapterInfo)GetProcAddress(gdi32_dll, "D3DKMTQueryAdapterInfo");
+        (PFN_LoaderQueryAdapterInfo)(void *)GetProcAddress(gdi32_dll, "D3DKMTQueryAdapterInfo");
     if (fpLoaderEnumAdapters2 == NULL || fpLoaderQueryAdapterInfo == NULL) {
         result = VK_ERROR_INCOMPATIBLE_DRIVER;
         goto out;
@@ -615,7 +622,8 @@ VkResult windows_read_manifest_from_d3d_adapters(const struct loader_instance *i
             .value_type = REG_MULTI_SZ,
             .physical_adapter_index = 0,
         };
-        wcsncpy(filename_info.value_name, value_name, sizeof(filename_info.value_name) / sizeof(WCHAR));
+        size_t value_name_size = wcslen(value_name);
+        wcsncpy_s(filename_info.value_name, MAX_PATH, value_name, value_name_size);
         LoaderQueryAdapterInfo query_info;
         query_info.handle = adapters.adapters[i].handle;
         query_info.type = LOADER_QUERY_TYPE_REGISTRY;
@@ -680,7 +688,7 @@ VkResult windows_read_manifest_from_d3d_adapters(const struct loader_instance *i
                 goto out;
             }
 
-            // If this is a string and not a multi-string, we don't want to go throught the loop more than once
+            // If this is a string and not a multi-string, we don't want to go through the loop more than once
             if (full_info->value_type == REG_SZ) {
                 break;
             }
@@ -698,17 +706,19 @@ out:
 // Look for data files in the registry.
 VkResult windows_read_data_files_in_registry(const struct loader_instance *inst, enum loader_data_files_type data_file_type,
                                              bool warn_if_not_present, char *registry_location,
-                                             struct loader_data_files *out_files) {
+                                             struct loader_string_list *out_files) {
     VkResult vk_result = VK_SUCCESS;
     char *search_path = NULL;
     uint32_t log_target_flag = 0;
 
     if (data_file_type == LOADER_DATA_FILE_MANIFEST_DRIVER) {
         log_target_flag = VULKAN_LOADER_DRIVER_BIT;
-        loader_log(inst, log_target_flag, 0, "Checking for Driver Manifest files in Registry at %s", registry_location);
+        loader_log(inst, log_target_flag, 0, "Checking for Driver Manifest files in Registry at %s\\%s",
+                   DEFAULT_VK_REGISTRY_HIVE_STR, registry_location);
     } else {
         log_target_flag = VULKAN_LOADER_LAYER_BIT;
-        loader_log(inst, log_target_flag, 0, "Checking for Layer Manifest files in Registry at %s", registry_location);
+        loader_log(inst, log_target_flag, 0, "Checking for Layer Manifest files in Registry at %s\\%s",
+                   DEFAULT_VK_REGISTRY_HIVE_STR, registry_location);
     }
 
     // These calls look at the PNP/Device section of the registry.
@@ -783,12 +793,143 @@ out:
     return vk_result;
 }
 
-// This function allocates an array in sorted_devices which must be freed by the caller if not null
-VkResult windows_read_sorted_physical_devices(struct loader_instance *inst, uint32_t *sorted_devices_count,
-                                              struct loader_phys_dev_per_icd **sorted_devices) {
+VkResult enumerate_adapter_physical_devices(struct loader_instance *inst, struct loader_icd_term *icd_term, LUID luid,
+                                            uint32_t *icd_phys_devs_array_count,
+                                            struct loader_icd_physical_devices *icd_phys_devs_array) {
+    uint32_t count = 0;
+    VkResult res = icd_term->scanned_icd->EnumerateAdapterPhysicalDevices(icd_term->instance, luid, &count, NULL);
+    if (res == VK_ERROR_OUT_OF_HOST_MEMORY) {
+        return res;
+    } else if (res == VK_ERROR_INCOMPATIBLE_DRIVER || res == VK_ERROR_INITIALIZATION_FAILED || 0 == count) {
+        return VK_SUCCESS;  // This driver doesn't support the adapter
+    } else if (res != VK_SUCCESS) {
+        loader_log(inst, VULKAN_LOADER_WARN_BIT, 0,
+                   "Failed to convert DXGI adapter into Vulkan physical device with unexpected error code: %d", res);
+        return VK_SUCCESS;
+    }
+
+    // Take a pointer to the last element of icd_phys_devs_array to simplify usage
+    struct loader_icd_physical_devices *next_icd_phys_devs = &icd_phys_devs_array[*icd_phys_devs_array_count];
+
+    // Get the actual physical devices
+    do {
+        next_icd_phys_devs->physical_devices = loader_instance_heap_realloc(
+            inst, next_icd_phys_devs->physical_devices, next_icd_phys_devs->device_count * sizeof(VkPhysicalDevice),
+            count * sizeof(VkPhysicalDevice), VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);
+        if (next_icd_phys_devs->physical_devices == NULL) {
+            return VK_ERROR_OUT_OF_HOST_MEMORY;
+        }
+        next_icd_phys_devs->device_count = count;
+    } while ((res = icd_term->scanned_icd->EnumerateAdapterPhysicalDevices(icd_term->instance, luid, &count,
+                                                                           next_icd_phys_devs->physical_devices)) == VK_INCOMPLETE);
+
+    if (res != VK_SUCCESS) {
+        loader_instance_heap_free(inst, next_icd_phys_devs->physical_devices);
+        next_icd_phys_devs->physical_devices = NULL;
+        // Unless OOHM occurs, only return VK_SUCCESS
+        if (res != VK_ERROR_OUT_OF_HOST_MEMORY) {
+            res = VK_SUCCESS;
+            loader_log(inst, VULKAN_LOADER_WARN_BIT, 0, "Failed to convert DXGI adapter into Vulkan physical device");
+        }
+        return res;
+    }
+
+    // Because the loader calls EnumerateAdapterPhysicalDevices on all drivers with each DXGI Adapter, if there are multiple drivers
+    // that share a luid the physical device will get queried multiple times. We can prevent that by not adding them if the
+    // enumerated physical devices have already been added.
+    bool already_enumerated = false;
+    for (uint32_t j = 0; j < *icd_phys_devs_array_count; j++) {
+        if (count == icd_phys_devs_array[j].device_count) {
+            bool matches = true;
+            for (uint32_t k = 0; k < icd_phys_devs_array[j].device_count; k++) {
+                if (icd_phys_devs_array[j].physical_devices[k] != next_icd_phys_devs->physical_devices[k]) {
+                    matches = false;
+                    break;
+                }
+            }
+            if (matches) {
+                already_enumerated = true;
+            }
+        }
+    }
+    if (!already_enumerated) {
+        next_icd_phys_devs->device_count = count;
+        next_icd_phys_devs->icd_term = icd_term;
+        next_icd_phys_devs->windows_adapter_luid = luid;
+        (*icd_phys_devs_array_count)++;
+    } else {
+        // Avoid memory leak in case of the already_enumerated hitting true
+        // at the last enumerate_adapter_physical_devices call in the outer loop
+        loader_instance_heap_free(inst, next_icd_phys_devs->physical_devices);
+        next_icd_phys_devs->physical_devices = NULL;
+    }
+
+    return VK_SUCCESS;
+}
+
+// Whenever there are multiple drivers for the same hardware and one of the drivers is an implementation layered on top of another
+// API (such as the Dozen driver which converts vulkan to Dx12), we want to make sure the layered driver appears after the 'native'
+// driver. This function iterates over all physical devices and make sure any with matching LUID's are sorted such that drivers with
+// a underlyingAPI of VK_LAYERED_DRIVER_UNDERLYING_API_D3D12_MSFT are ordered after drivers without it.
+void sort_physical_devices_with_same_luid(struct loader_instance *inst, uint32_t icd_phys_devs_array_count,
+                                          struct loader_icd_physical_devices *icd_phys_devs_array) {
+    bool app_is_vulkan_1_1 = loader_check_version_meets_required(LOADER_VERSION_1_1_0, inst->app_api_version);
+
+    for (uint32_t i = 0; icd_phys_devs_array_count > 1 && i < icd_phys_devs_array_count - 1; i++) {
+        for (uint32_t j = i + 1; j < icd_phys_devs_array_count; j++) {
+            // Only want to reorder physical devices if their ICD's LUID's match
+            if ((icd_phys_devs_array[i].windows_adapter_luid.HighPart != icd_phys_devs_array[j].windows_adapter_luid.HighPart) ||
+                (icd_phys_devs_array[i].windows_adapter_luid.LowPart != icd_phys_devs_array[j].windows_adapter_luid.LowPart)) {
+                continue;
+            }
+
+            VkLayeredDriverUnderlyingApiMSFT underlyingAPI = VK_LAYERED_DRIVER_UNDERLYING_API_NONE_MSFT;
+            VkPhysicalDeviceLayeredDriverPropertiesMSFT layered_driver_properties_msft = {0};
+            layered_driver_properties_msft.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LAYERED_DRIVER_PROPERTIES_MSFT;
+            VkPhysicalDeviceProperties2 props2 = {0};
+            props2.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROPERTIES_2;
+            props2.pNext = (void *)&layered_driver_properties_msft;
+
+            // Because there may be multiple physical devices associated with each ICD, we need to check each physical device
+            // whether it is layered
+            for (uint32_t k = 0; k < icd_phys_devs_array[i].device_count; k++) {
+                VkPhysicalDeviceProperties dev_props = {0};
+                icd_phys_devs_array[i].icd_term->dispatch.GetPhysicalDeviceProperties(icd_phys_devs_array[i].physical_devices[k],
+                                                                                      &dev_props);
+
+                bool device_is_1_1_capable =
+                    loader_check_version_meets_required(LOADER_VERSION_1_1_0, loader_make_version(dev_props.apiVersion));
+
+                PFN_vkGetPhysicalDeviceProperties2 GetPhysDevProps2 = NULL;
+                if (app_is_vulkan_1_1 && device_is_1_1_capable) {
+                    GetPhysDevProps2 = icd_phys_devs_array[i].icd_term->dispatch.GetPhysicalDeviceProperties2;
+                } else {
+                    GetPhysDevProps2 = (PFN_vkGetPhysicalDeviceProperties2)icd_phys_devs_array[i]
+                                           .icd_term->dispatch.GetPhysicalDeviceProperties2KHR;
+                }
+                if (GetPhysDevProps2) {
+                    GetPhysDevProps2(icd_phys_devs_array[i].physical_devices[k], &props2);
+                    if (layered_driver_properties_msft.underlyingAPI != VK_LAYERED_DRIVER_UNDERLYING_API_NONE_MSFT) {
+                        underlyingAPI = layered_driver_properties_msft.underlyingAPI;
+                        break;
+                    }
+                }
+            }
+            if (underlyingAPI == VK_LAYERED_DRIVER_UNDERLYING_API_D3D12_MSFT) {
+                struct loader_icd_physical_devices swap_icd = icd_phys_devs_array[i];
+                icd_phys_devs_array[i] = icd_phys_devs_array[j];
+                icd_phys_devs_array[j] = swap_icd;
+            }
+        }
+    }
+}
+
+// This function allocates icd_phys_devs_array which must be freed by the caller if not null
+VkResult windows_read_sorted_physical_devices(struct loader_instance *inst, uint32_t *icd_phys_devs_array_count,
+                                              struct loader_icd_physical_devices **icd_phys_devs_array) {
     VkResult res = VK_SUCCESS;
 
-    uint32_t sorted_alloc = 0;
+    uint32_t icd_phys_devs_array_size = 0;
     struct loader_icd_term *icd_term = NULL;
     IDXGIFactory6 *dxgi_factory = NULL;
     HRESULT hres = fpCreateDXGIFactory1(&IID_IDXGIFactory6, (void **)&dxgi_factory);
@@ -796,10 +937,10 @@ VkResult windows_read_sorted_physical_devices(struct loader_instance *inst, uint
         loader_log(inst, VULKAN_LOADER_INFO_BIT, 0, "Failed to create DXGI factory 6. Physical devices will not be sorted");
         goto out;
     }
-    sorted_alloc = 16;
-    *sorted_devices = loader_instance_heap_calloc(inst, sorted_alloc * sizeof(struct loader_phys_dev_per_icd),
-                                                  VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);
-    if (*sorted_devices == NULL) {
+    icd_phys_devs_array_size = 16;
+    *icd_phys_devs_array = loader_instance_heap_calloc(inst, icd_phys_devs_array_size * sizeof(struct loader_icd_physical_devices),
+                                                       VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);
+    if (*icd_phys_devs_array == NULL) {
         res = VK_ERROR_OUT_OF_HOST_MEMORY;
         goto out;
     }
@@ -823,74 +964,35 @@ VkResult windows_read_sorted_physical_devices(struct loader_instance *inst, uint
             continue;
         }
 
-        if (sorted_alloc <= i) {
-            uint32_t old_size = sorted_alloc * sizeof(struct loader_phys_dev_per_icd);
-            *sorted_devices =
-                loader_instance_heap_realloc(inst, *sorted_devices, old_size, 2 * old_size, VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);
-            if (*sorted_devices == NULL) {
+        if (icd_phys_devs_array_size <= i) {
+            uint32_t old_size = icd_phys_devs_array_size * sizeof(struct loader_icd_physical_devices);
+            *icd_phys_devs_array = loader_instance_heap_realloc(inst, *icd_phys_devs_array, old_size, 2 * old_size,
+                                                                VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);
+            if (*icd_phys_devs_array == NULL) {
                 adapter->lpVtbl->Release(adapter);
                 res = VK_ERROR_OUT_OF_HOST_MEMORY;
                 goto out;
             }
-            sorted_alloc *= 2;
+            icd_phys_devs_array_size *= 2;
         }
-        struct loader_phys_dev_per_icd *sorted_array = *sorted_devices;
-        sorted_array[*sorted_devices_count].device_count = 0;
-        sorted_array[*sorted_devices_count].physical_devices = NULL;
+        (*icd_phys_devs_array)[*icd_phys_devs_array_count].device_count = 0;
+        (*icd_phys_devs_array)[*icd_phys_devs_array_count].physical_devices = NULL;
 
         icd_term = inst->icd_terms;
-        for (uint32_t icd_idx = 0; NULL != icd_term; icd_term = icd_term->next, icd_idx++) {
+        while (NULL != icd_term) {
             // This is the new behavior, which cannot be run unless the ICD provides EnumerateAdapterPhysicalDevices
             if (icd_term->scanned_icd->EnumerateAdapterPhysicalDevices == NULL) {
+                icd_term = icd_term->next;
                 continue;
             }
 
-            uint32_t count = 0;
-            VkResult vkres =
-                icd_term->scanned_icd->EnumerateAdapterPhysicalDevices(icd_term->instance, description.AdapterLuid, &count, NULL);
-            if (vkres == VK_ERROR_INCOMPATIBLE_DRIVER) {
-                continue;  // This driver doesn't support the adapter
-            } else if (vkres == VK_ERROR_OUT_OF_HOST_MEMORY) {
-                res = VK_ERROR_OUT_OF_HOST_MEMORY;
+            res = enumerate_adapter_physical_devices(inst, icd_term, description.AdapterLuid, icd_phys_devs_array_count,
+                                                     *icd_phys_devs_array);
+            if (res == VK_ERROR_OUT_OF_HOST_MEMORY) {
+                adapter->lpVtbl->Release(adapter);
                 goto out;
-            } else if (vkres != VK_SUCCESS) {
-                loader_log(inst, VULKAN_LOADER_WARN_BIT, 0,
-                           "Failed to convert DXGI adapter into Vulkan physical device with unexpected error code");
-                continue;
             }
-
-            // Get the actual physical devices
-            if (0 != count) {
-                do {
-                    sorted_array[*sorted_devices_count].physical_devices =
-                        loader_instance_heap_realloc(inst, sorted_array[*sorted_devices_count].physical_devices,
-                                                     sorted_array[*sorted_devices_count].device_count * sizeof(VkPhysicalDevice),
-                                                     count * sizeof(VkPhysicalDevice), VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);
-                    if (sorted_array[*sorted_devices_count].physical_devices == NULL) {
-                        res = VK_ERROR_OUT_OF_HOST_MEMORY;
-                        break;
-                    }
-                    sorted_array[*sorted_devices_count].device_count = count;
-                } while ((vkres = icd_term->scanned_icd->EnumerateAdapterPhysicalDevices(
-                              icd_term->instance, description.AdapterLuid, &count,
-                              sorted_array[*sorted_devices_count].physical_devices)) == VK_INCOMPLETE);
-            }
-
-            if (vkres != VK_SUCCESS) {
-                loader_instance_heap_free(inst, sorted_array[*sorted_devices_count].physical_devices);
-                sorted_array[*sorted_devices_count].physical_devices = NULL;
-                if (vkres == VK_ERROR_OUT_OF_HOST_MEMORY) {
-                    res = VK_ERROR_OUT_OF_HOST_MEMORY;
-                    goto out;
-                } else {
-                    loader_log(inst, VULKAN_LOADER_WARN_BIT, 0, "Failed to convert DXGI adapter into Vulkan physical device");
-                    continue;
-                }
-            }
-            sorted_array[*sorted_devices_count].device_count = count;
-            sorted_array[*sorted_devices_count].icd_index = icd_idx;
-            sorted_array[*sorted_devices_count].icd_term = icd_term;
-            (*sorted_devices_count)++;
+            icd_term = icd_term->next;
         }
 
         adapter->lpVtbl->Release(adapter);
@@ -898,13 +1000,13 @@ VkResult windows_read_sorted_physical_devices(struct loader_instance *inst, uint
 
     dxgi_factory->lpVtbl->Release(dxgi_factory);
 
+    sort_physical_devices_with_same_luid(inst, *icd_phys_devs_array_count, *icd_phys_devs_array);
+
 out:
-    if (*sorted_devices_count == 0 && *sorted_devices != NULL) {
-        loader_instance_heap_free(inst, *sorted_devices);
-        *sorted_devices = NULL;
+    if (*icd_phys_devs_array_count == 0 && *icd_phys_devs_array != NULL) {
+        loader_instance_heap_free(inst, *icd_phys_devs_array);
+        *icd_phys_devs_array = NULL;
     }
-    *sorted_devices_count = *sorted_devices_count;
-    *sorted_devices = *sorted_devices;
     return res;
 }
 
@@ -923,7 +1025,7 @@ VkLoaderFeatureFlags windows_initialize_dxgi(void) {
 // Multiple groups could have devices out of the same sorted list, however, a single group's devices must all come
 // from the same sorted list.
 void windows_sort_devices_in_group(struct loader_instance *inst, struct VkPhysicalDeviceGroupProperties *group_props,
-                                   struct loader_phys_dev_per_icd *icd_sorted_list) {
+                                   struct loader_icd_physical_devices *icd_sorted_list) {
     uint32_t cur_index = 0;
     for (uint32_t dev = 0; dev < icd_sorted_list->device_count; ++dev) {
         for (uint32_t grp_dev = cur_index; grp_dev < group_props->physicalDeviceCount; ++grp_dev) {
@@ -948,7 +1050,7 @@ void windows_sort_devices_in_group(struct loader_instance *inst, struct VkPhysic
 VkResult windows_sort_physical_device_groups(struct loader_instance *inst, const uint32_t group_count,
                                              struct loader_physical_device_group_term *sorted_group_term,
                                              const uint32_t sorted_device_count,
-                                             struct loader_phys_dev_per_icd *sorted_phys_dev_array) {
+                                             struct loader_icd_physical_devices *sorted_phys_dev_array) {
     if (0 == group_count || NULL == sorted_group_term) {
         loader_log(inst, VULKAN_LOADER_WARN_BIT, 0,
                    "windows_sort_physical_device_groups: Called with invalid information (Group count %d, Sorted Info %p)",
@@ -993,27 +1095,25 @@ char *windows_get_app_package_manifest_path(const struct loader_instance *inst)
     // These functions are only available on Windows 8 and above, load them dynamically for compatibility with Windows 7
     typedef LONG(WINAPI * PFN_GetPackagesByPackageFamily)(PCWSTR, UINT32 *, PWSTR *, UINT32 *, WCHAR *);
     PFN_GetPackagesByPackageFamily fpGetPackagesByPackageFamily =
-        (PFN_GetPackagesByPackageFamily)GetProcAddress(GetModuleHandle(TEXT("kernel32.dll")), "GetPackagesByPackageFamily");
+        (PFN_GetPackagesByPackageFamily)(void *)GetProcAddress(GetModuleHandle(TEXT("kernel32.dll")), "GetPackagesByPackageFamily");
     if (!fpGetPackagesByPackageFamily) {
         return NULL;
     }
     typedef LONG(WINAPI * PFN_GetPackagePathByFullName)(PCWSTR, UINT32 *, PWSTR);
     PFN_GetPackagePathByFullName fpGetPackagePathByFullName =
-        (PFN_GetPackagePathByFullName)GetProcAddress(GetModuleHandle(TEXT("kernel32.dll")), "GetPackagePathByFullName");
+        (PFN_GetPackagePathByFullName)(void *)GetProcAddress(GetModuleHandle(TEXT("kernel32.dll")), "GetPackagePathByFullName");
     if (!fpGetPackagePathByFullName) {
         return NULL;
     }
 
     UINT32 numPackages = 0, bufferLength = 0;
-    /* This literal string identifies the Microsoft-published OpenCL and OpenGL Compatibility Pack
-     * (so named at the time this is being added), which contains OpenGLOn12 and OpenCLOn12 mapping
-     * layers, and will contain VulkanOn12 (aka Dozen) going forward.
-     */
+    // This literal string identifies the Microsoft-published OpenCL, OpenGL, and Vulkan Compatibility Pack, which contains
+    // OpenGLOn12, OpenCLOn12, and VulkanOn12 (aka Dozen) mappinglayers
     PCWSTR familyName = L"Microsoft.D3DMappingLayers_8wekyb3d8bbwe";
     if (ERROR_INSUFFICIENT_BUFFER != fpGetPackagesByPackageFamily(familyName, &numPackages, NULL, &bufferLength, NULL) ||
         numPackages == 0 || bufferLength == 0) {
         loader_log(inst, VULKAN_LOADER_INFO_BIT, 0,
-                   "windows_get_app_package_manifest_path: Failed to find mapping layers packages by family name\n");
+                   "windows_get_app_package_manifest_path: Failed to find mapping layers packages by family name");
         return NULL;
     }
 
@@ -1022,13 +1122,13 @@ char *windows_get_app_package_manifest_path(const struct loader_instance *inst)
     PWSTR *packages = loader_instance_heap_alloc(inst, sizeof(PWSTR) * numPackages, VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);
     if (!buffer || !packages) {
         loader_log(inst, VULKAN_LOADER_ERROR_BIT, 0,
-                   "windows_get_app_package_manifest_path: Failed to allocate memory for package names\n");
+                   "windows_get_app_package_manifest_path: Failed to allocate memory for package names");
         goto cleanup;
     }
 
     if (ERROR_SUCCESS != fpGetPackagesByPackageFamily(familyName, &numPackages, packages, &bufferLength, buffer)) {
         loader_log(inst, VULKAN_LOADER_ERROR_BIT, 0,
-                   "windows_get_app_package_manifest_path: Failed to mapping layers package full names\n");
+                   "windows_get_app_package_manifest_path: Failed to mapping layers package full names");
         goto cleanup;
     }
 
@@ -1038,20 +1138,20 @@ char *windows_get_app_package_manifest_path(const struct loader_instance *inst)
     if (ERROR_INSUFFICIENT_BUFFER != fpGetPackagePathByFullName(packages[0], &pathLength, NULL) || pathLength > MAX_PATH ||
         ERROR_SUCCESS != fpGetPackagePathByFullName(packages[0], &pathLength, path)) {
         loader_log(inst, VULKAN_LOADER_ERROR_BIT, 0,
-                   "windows_get_app_package_manifest_path: Failed to get mapping layers package path\n");
+                   "windows_get_app_package_manifest_path: Failed to get mapping layers package path");
         goto cleanup;
     }
 
     int narrowPathLength = WideCharToMultiByte(CP_ACP, 0, path, -1, NULL, 0, NULL, NULL);
     if (narrowPathLength == 0) {
         loader_log(inst, VULKAN_LOADER_ERROR_BIT, 0,
-                   "windows_get_app_package_manifest_path: Failed to convert path from wide to narrow\n");
+                   "windows_get_app_package_manifest_path: Failed to convert path from wide to narrow");
         goto cleanup;
     }
 
     ret = loader_instance_heap_alloc(inst, narrowPathLength, VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);
     if (!ret) {
-        loader_log(inst, VULKAN_LOADER_ERROR_BIT, 0, "windows_get_app_package_manifest_path: Failed to allocate path\n");
+        loader_log(inst, VULKAN_LOADER_ERROR_BIT, 0, "windows_get_app_package_manifest_path: Failed to allocate path");
         goto cleanup;
     }
 
@@ -1063,4 +1163,95 @@ cleanup:
     loader_instance_heap_free(inst, packages);
     return ret;
 }
+
+VkResult get_settings_path_if_exists_in_registry_key(const struct loader_instance *inst, char **out_path, HKEY key) {
+    VkResult result = VK_ERROR_INITIALIZATION_FAILED;
+
+    char name[MAX_STRING_SIZE] = {0};
+    DWORD name_size = sizeof(name);
+
+    *out_path = NULL;
+
+    LONG rtn_value = ERROR_SUCCESS;
+    for (DWORD idx = 0; rtn_value == ERROR_SUCCESS; idx++) {
+        DWORD value = 0;
+        DWORD value_size = sizeof(value);
+        rtn_value = RegEnumValue(key, idx, name, &name_size, NULL, NULL, (LPBYTE)&value, &value_size);
+
+        if (ERROR_SUCCESS != rtn_value) {
+            break;
+        }
+
+        uint32_t start_of_path_filename = 0;
+        for (uint32_t last_char = name_size; last_char > 0; last_char--) {
+            if (name[last_char] == '\\') {
+                start_of_path_filename = last_char + 1;
+                break;
+            }
+        }
+
+        // Make sure the path exists first
+        if (*out_path && !loader_platform_file_exists(name)) {
+            return VK_ERROR_INITIALIZATION_FAILED;
+        }
+
+        if (strcmp(VK_LOADER_SETTINGS_FILENAME, &(name[start_of_path_filename])) == 0) {
+            *out_path = loader_instance_heap_calloc(inst, name_size + 1, VK_SYSTEM_ALLOCATION_SCOPE_INSTANCE);
+            if (*out_path == NULL) {
+                return VK_ERROR_OUT_OF_HOST_MEMORY;
+            }
+            loader_strncpy(*out_path, name_size + 1, name, name_size);
+            (*out_path)[name_size] = '\0';
+            result = VK_SUCCESS;
+            break;
+        }
+    }
+
+    return result;
+}
+
+VkResult windows_get_loader_settings_file_path(const struct loader_instance *inst, char **out_path) {
+    VkResult result = VK_SUCCESS;
+    DWORD access_flags = KEY_QUERY_VALUE;
+    LONG rtn_value = 0;
+    HKEY key = NULL;
+    *out_path = NULL;
+
+    // Search in HKEY_CURRENT_USER first if we are running without admin privileges
+    // Exit if a settings file was found.
+    // Otherwise check in HKEY_LOCAL_MACHINE.
+
+    if (!is_high_integrity()) {
+        rtn_value = RegOpenKeyEx(HKEY_CURRENT_USER, VK_SETTINGS_INFO_REGISTRY_LOC, 0, access_flags, &key);
+        if (ERROR_SUCCESS == rtn_value) {
+            result = get_settings_path_if_exists_in_registry_key(inst, out_path, key);
+
+            // Either we got OOM and *must* exit or we successfully found the settings file and can exit
+            if (result == VK_ERROR_OUT_OF_HOST_MEMORY || result == VK_SUCCESS) {
+                goto out;
+            }
+            RegCloseKey(key);
+            key = NULL;
+        }
+    }
+
+    rtn_value = RegOpenKeyEx(HKEY_LOCAL_MACHINE, VK_SETTINGS_INFO_REGISTRY_LOC, 0, access_flags, &key);
+    if (ERROR_SUCCESS != rtn_value) {
+        result = VK_ERROR_FEATURE_NOT_PRESENT;
+        goto out;
+    }
+
+    result = get_settings_path_if_exists_in_registry_key(inst, out_path, key);
+    if (result == VK_ERROR_OUT_OF_HOST_MEMORY) {
+        goto out;
+    }
+
+out:
+    if (NULL != key) {
+        RegCloseKey(key);
+    }
+
+    return result;
+}
+
 #endif  // _WIN32
